import math
from datetime import datetime, timedelta
from dateutil.tz import tzutc
from bisect import bisect_left, bisect_right

from pscr_common.aws import get_boto3_client
from pscr_common.health_check import update_health_check
import logging
import time
import bisect
from org_mapper.utils import log_and_sleep
from pyformance import time_calls, global_registry

_logger = logging.getLogger(__name__)


class KinesisMaxRetriesException(Exception):
    def __init__(self, shard_id, last_logged_exception, message="Max retries exceeded"):
        self.message = f"{message} for shard_id: {shard_id} \nLast logged exception: {last_logged_exception}"
        super().__init__(self.message)


class ShardManager:
    kinesis_client = get_boto3_client("kinesis")
    s3_client = get_boto3_client("s3")
    sts_client = get_boto3_client("sts")

    def __init__(self, partition_keys, partition_metadata, config):
        self.partition_keys = partition_keys
        self.partition_metadata = partition_metadata
        self.stream_name = config.get("aws", "kinesis_stream_name")
        self.min_shard_util = config.getfloat("shard_management", "min_shard_util")
        self.max_shard_util = config.getfloat("shard_management", "max_shard_util")
        self.shard_splits = []
        self.shard_merges = []
        self.initial_active_shards = None
        self.reshard_passes = config.getint("shard_management", "reshard_passes")
        self.reshard_operations_limit = config.getint(
            "shard_management", "reshard_operations_limit"
        )
        self.reshard_by_priority = config.getboolean(
            "shard_management", "reshard_by_priority"
        )
        self.default_partition_priority = config.getint(
            "shard_management", "default_partition_priority"
        )
        self.reshard_operations = 0
        self.reshard_operations_low_priority = 0
        self.xdre_replay_enabled = config.getboolean(
            "shard_management", "xdre_replay_enabled"
        )
        self.reshard_operations_low_priority_limit = config.getint(
            "shard_management", "reshard_operations_low_priority_limit"
        )
        self.split_shard_in_middle = config.getboolean(
            "shard_management", "split_shard_in_middle"
        )
        self.merge_shards_by_pairs = config.getboolean(
            "shard_management", "merge_shards_by_pairs"
        )
        self.shard_settle_timer = config.getint(
            "shard_management", "shard_settle_timer"
        )
        self.new_shards_map = []
        self.split_pass_ran = False
        self.max_kinesis_client_retries = config.getint(
            "shard_management", "max_kinesis_client_retries"
        )
        self.max_shards_already_split_or_merged = config.getint(
            "shard_management", "max_shards_already_split_or_merged"
        )
        self.max_partition_keys_per_shard = config.getint(
            "shard_management", "max_partition_keys_per_shard"
        )
        self.min_partition_keys_per_unutilized_shard = config.getint(
            "shard_management", "min_partition_keys_per_unutilized_shard"
        )
        self.avg_partition_keys_per_shard = math.ceil(
            (
                self.max_partition_keys_per_shard
                + self.min_partition_keys_per_unutilized_shard
            )
            / 2
        )
        self.count_shards_already_split_or_merged = 0
        self.metrics_registry = global_registry()
        self.assume_role = config.getboolean("shard_management", "assume_role")
        self.role_arn = config.get("shard_management", "role_arn")
        self.role_expiration = datetime.now(tzutc())
        self.env_id = config.get("env", "id")
        self.mode = config.get("service", "mode")

    def _assume_role(self):
        # Override the kinesis_client with one in an assumed role context, to access kinesis stream in a different
        # aws account.
        if self.assume_role:
            # we have to re-up our assume role every hour, we just check if we are close to the deadline before making
            # kinesis API calls.
            if self.role_expiration - timedelta(minutes=2) < datetime.now(tzutc()):
                _logger.info(
                    f"Assuming role for kinesis stream {self.role_arn}, Stream Name {self.stream_name}"
                )
                role_session_name = f"{self.env_id}-{self.mode}-ShardManagerKinesisRole"
                assume_role_object = self.sts_client.assume_role(
                    RoleArn=self.role_arn,
                    RoleSessionName=role_session_name,
                    DurationSeconds=3600,
                )
                credentials = assume_role_object["Credentials"]
                self.role_expiration = credentials["Expiration"]
                self.kinesis_client = get_boto3_client(
                    "kinesis",
                    aws_access_key_id=credentials["AccessKeyId"],
                    aws_secret_access_key=credentials["SecretAccessKey"],
                    aws_session_token=credentials["SessionToken"],
                )

    def set_initial_active_shards(self):
        self.initial_active_shards = self.get_active_shards()

    @time_calls
    def build_new_stream_shards(self):
        update_health_check()
        if sum(self.partition_keys.values()) == 0:
            # if we don't have any utilized partition keys (maybe due to no CW data, or the data shows no traffic),
            # return here without computing a new_shards_map (prompts org mapper to skip resharding)
            _logger.warning(
                "No partition keys are utilized (possibly due to CloudWatch metrics issue). "
                "Returning without computing a new_shards_map."
            )
            return

        stream_shards = self._get_current_stream_shards()
        not_finished = True
        prev_shard = None
        index = 0
        while not_finished:
            #  Each time we split or modify a shard we break out of this for loop
            #  So that we can re-process the same index and continue the calculations with the new shards
            #  stream_shards is modified on the fly in this loop
            for shard, next_shard in self._pairwise(stream_shards, index):
                shard_percents = self._get_partition_keys_in_shard(shard)
                shard_percent_util = sum(shard_percents.values())

                if (
                    shard_percent_util > self.max_shard_util
                    or 0 < self.max_partition_keys_per_shard < len(shard_percents)
                ):
                    stream_shards = self._split_virtual_shard(shard, stream_shards)
                    not_finished = True
                    break
                elif shard_percent_util < self.min_shard_util and (
                    len(shard_percents) < self.min_partition_keys_per_unutilized_shard
                    or self.min_partition_keys_per_unutilized_shard == 0
                ):
                    if next_shard is not None:
                        partition_keys_in_shard = self._get_partition_keys_in_shard(
                            shard
                        )
                        partition_keys_in_next_shard = (
                            self._get_partition_keys_in_shard(next_shard)
                        )
                        if len(partition_keys_in_next_shard) == 0:
                            next_key_bandwidth = 0
                        else:
                            next_key_bandwidth = partition_keys_in_next_shard[
                                sorted(partition_keys_in_next_shard.keys())[0]
                            ]
                        # Merge Any Empty shards found or 0 utilized shards
                        if len(partition_keys_in_shard) == 0 or shard_percent_util == 0:
                            if len(partition_keys_in_shard) == 0:
                                _logger.debug(
                                    f"Found an empty shard: {len(partition_keys_in_shard)}, "
                                    f"next_shard {len(partition_keys_in_next_shard)}"
                                )
                            else:
                                _logger.debug(
                                    f"Found shard with {shard_percent_util} utilization"
                                )
                            stream_shards = self._merge_virtual_shard(
                                shard, next_shard, stream_shards
                            )
                            not_finished = True
                            break
                        # Don't merge if the next partition key will take it over 100 percent
                        if shard_percent_util + next_key_bandwidth < 1 and (
                            0 == self.max_partition_keys_per_shard
                            or len(shard_percents) + len(partition_keys_in_next_shard)
                            < self.max_partition_keys_per_shard
                        ):
                            stream_shards = self._merge_virtual_shard(
                                shard, next_shard, stream_shards
                            )
                            not_finished = True
                            break
                # Special check for last shard to be empty and merge it if so
                if (
                    next_shard is None
                    and len(self._get_partition_keys_in_shard(shard)) == 0
                    and prev_shard is not None
                ):
                    _logger.info(f"Last shard is empty merging with next to last shard")
                    stream_shards = self._merge_virtual_shard(
                        prev_shard, shard, stream_shards
                    )
                prev_shard = shard
                shard[2] = shard_percent_util
                not_finished = False
                index += 1

        if self.reshard_by_priority:
            self.new_shards_map = self._sort_stream_shards_by_priority(stream_shards)
        else:
            self.new_shards_map = self._sort_stream_shards(stream_shards)

    def _split_virtual_shard(self, shard, stream_shards):
        shard_partition_keys = self._get_partition_keys_in_shard(shard)
        shard_bandwidth_percent = 0
        shard_partition_keys_count = 0
        next_key = None
        for current_key, next_key in self._pairwise(
            sorted(shard_partition_keys.keys())
        ):
            shard_bandwidth_percent += shard_partition_keys[current_key]
            shard_partition_keys_count += 1
            if next_key is None:
                next_bandwidth = 0
            else:
                next_bandwidth = shard_bandwidth_percent + self.partition_keys[next_key]
            if (
                next_bandwidth > 1
                and shard_bandwidth_percent != 0
                or (0 < self.avg_partition_keys_per_shard <= shard_partition_keys_count)
            ):
                _logger.debug("Shard got to 100 percent")
                partition_keys = self._get_partition_keys_in_shard(
                    [shard[0], current_key, shard_bandwidth_percent]
                )
                partition_priority = self._get_max_partition_priority(
                    partition_keys.keys()
                )
                stream_shards.append(
                    [shard[0], current_key, shard_bandwidth_percent, partition_priority]
                )
                _logger.debug(
                    f"First Shard Bandwidth: {shard_bandwidth_percent}, Start Hash: {shard[0]}, "
                    f"End Hash: {current_key}, "
                    f"Partition keys: "
                    f"{partition_keys}, "
                    f"Priority: {partition_priority}"
                )
                next_shard_bandwidth = self._get_shard_util(
                    [current_key + 1, shard[1], None, partition_priority]
                )
                partition_keys_2 = self._get_partition_keys_in_shard(
                    [current_key + 1, shard[1], next_shard_bandwidth]
                )
                partition_priority_2 = self._get_max_partition_priority(
                    partition_keys_2.keys()
                )
                stream_shards.append(
                    [
                        current_key + 1,
                        shard[1],
                        next_shard_bandwidth,
                        partition_priority_2,
                    ]
                )
                _logger.debug(
                    f"Second Shard Bandwidth: {next_shard_bandwidth}, Start Hash: {current_key + 1}, "
                    f"End Hash: {shard[1]}, "
                    f"Partition keys: "
                    f"{partition_keys_2}, "
                    f"Priority: {partition_priority_2}"
                )
                stream_shards.remove(shard)
                break
        if next_key is None:
            _logger.warning(
                f"We failed to split this shard when asked to:  {shard}, "
                f"Partition keys: {shard_partition_keys}"
            )
        return self._sort_stream_shards(stream_shards)

    def _merge_virtual_shard(self, shard, next_shard, stream_shards):
        try:
            shards = self._get_partition_keys_in_shard(
                [shard[0], next_shard[1] - 1, None]
            )
            shard_bandwidth = sum(shards.values())
            shard_priority = self._get_max_partition_priority(shards.keys())
        except TypeError:
            raise Exception(
                "Attempted to merge a non existent shard, likely only two shards left to merge together, "
                "could be a metrics input problem where we are trying to merge all shards together."
                "Exiting without performing any merge or split operations this run"
            )
        stream_shards.append([shard[0], next_shard[1], shard_bandwidth, shard_priority])
        stream_shards.remove(shard)
        stream_shards.remove(next_shard)
        return self._sort_stream_shards(stream_shards)

    def _get_current_stream_shards(self):
        #  Get current shards and format as a list of tuples to make re-shard calc easy.  Default percent full to None
        shards = self.get_active_shards()
        current_stream_shards = []
        for shard in shards:
            current_stream_shards.append(
                [
                    int(shard["HashKeyRange"]["StartingHashKey"]),
                    int(shard["HashKeyRange"]["EndingHashKey"]),
                    None,
                    self.default_partition_priority,
                ]
            )
        return self._sort_stream_shards(current_stream_shards)

    @staticmethod
    def _sort_stream_shards(stream_shards):
        new_stream_shards = []
        for shard in sorted(stream_shards, key=lambda i: i[0]):
            new_stream_shards.append(shard)
        return new_stream_shards

    def _get_shard_util(self, stream_shard):
        percent_util = []
        for partition_key in self.partition_keys:
            if stream_shard[0] <= partition_key <= stream_shard[1]:
                percent_util.append(self.partition_keys[partition_key])
        return sum(percent_util)

    def _get_partition_keys_in_shard(self, stream_shard):
        shard_partition_keys = {}
        for partition_key in self.partition_keys:
            if stream_shard[0] <= partition_key <= stream_shard[1]:
                shard_partition_keys[partition_key] = self.partition_keys[partition_key]
        return shard_partition_keys

    @staticmethod
    def _pairwise(lst, start=0):
        # Iterator to return current item and next item in a list.  Provides one item look ahead effectively.
        for i in range(start, len(lst) - 1):
            yield lst[i], lst[i + 1]
        yield lst[-1], None

    @staticmethod
    def _find_new_end_hash_middle(new_shard_map, start_hash, end_hash):
        shards_in_range = [
            shard
            for shard in new_shard_map
            if start_hash <= shard[1] and end_hash >= shard[0]
        ]
        shards_in_range = ShardManager._sort_stream_shards(shards_in_range)
        _logger.debug(
            f"Shards between {start_hash} and {end_hash} are: {shards_in_range}"
        )
        middle = ((len(shards_in_range) + 1) // 2) - 1
        _logger.debug(
            f"Shards between {start_hash} and {end_hash} having middle element {middle}  are: {shards_in_range}"
        )
        return shards_in_range[middle][1]

    @time_calls
    def get_shards(self):
        update_health_check()
        shards = []
        kwargs = {"StreamName": self.stream_name}
        self._assume_role()
        while True:
            retries = 0
            while True:
                try:
                    resp = self.kinesis_client.list_shards(**kwargs)
                except self.kinesis_client.exceptions.LimitExceededException:
                    log_and_sleep(
                        "Kinesis Limit Exceeded Exception retrying",
                        retries,
                        max_sleep=60,
                    )
                except Exception as e:
                    log_and_sleep(str(e), retries, max_sleep=60)
                else:
                    break
                retries += 1
            shards += resp["Shards"]
            try:
                kwargs = {"NextToken": resp["NextToken"]}
            except KeyError:
                break
        return shards

    def get_active_shards(self):
        return [
            shard
            for shard in self.get_shards()
            if "EndingSequenceNumber" not in shard["SequenceNumberRange"]
        ]

    def _check_shard_already_split_or_merged(self, shard_id, exception):
        """
        Kinesis InvalidArgumentException is sometimes caused by trying to split or merge a shard that has already been
        split or merged.
        This is an error which we can ignore in most cases and continue with the org mapper run uninterruptedly.
        If we start seeing it too often though, we raise an exception.
        """
        if "has already been merged or split" in str(exception):
            self.count_shards_already_split_or_merged += 1
            if (
                self.count_shards_already_split_or_merged
                >= self.max_shards_already_split_or_merged
            ):
                raise KinesisMaxRetriesException(
                    shard_id,
                    exception,
                    message="Threshold reached for shard already split or merged errors",
                )
            else:
                _logger.debug(str(exception))
                return True
        else:
            return False

    @time_calls
    def _split_shard(self, shard_id, shard_split_hash):
        update_health_check()
        self.shard_splits.append((shard_id, shard_split_hash))
        kwargs = {
            "StreamName": self.stream_name,
            "ShardToSplit": shard_id,
            "NewStartingHashKey": str(shard_split_hash),
        }
        retries = 0
        last_logged_exception = ""
        self._assume_role()
        while True:
            if retries > self.max_kinesis_client_retries:
                raise KinesisMaxRetriesException(
                    shard_id, last_logged_exception, "Shard splitting retries exceeded"
                )
            try:
                self.kinesis_client.split_shard(**kwargs)
            except self.kinesis_client.exceptions.LimitExceededException:
                last_logged_exception = log_and_sleep(
                    f"Kinesis Limit Exceeded Exception retrying shardid: "
                    f"{shard_id} split_hash: {shard_split_hash}",
                    retries,
                )
            except self.kinesis_client.exceptions.ResourceInUseException:
                last_logged_exception = log_and_sleep(
                    f"Kinesis Resource In Use Exception retrying shardid: "
                    f"{shard_id} split_hash: {shard_split_hash}",
                    retries,
                )
            except self.kinesis_client.exceptions.InvalidArgumentException as e:
                if self._check_shard_already_split_or_merged(shard_id, e):
                    # disregard if the shard has already been merged or split
                    break
                last_logged_exception = log_and_sleep(
                    f"Shard Splitting Error: {str(e)}", retries
                )
            except Exception as e:
                last_logged_exception = log_and_sleep(
                    f"{e} retrying shardid: {shard_id} "
                    f"split_hash: {shard_split_hash}",
                    retries,
                )
            else:
                break

            retries += 1

    @time_calls
    def _merge_shards(self, shard_id, adjacent_shard_id):
        update_health_check()
        self.shard_merges.append((shard_id, adjacent_shard_id))
        kwargs = {
            "StreamName": self.stream_name,
            "ShardToMerge": shard_id,
            "AdjacentShardToMerge": adjacent_shard_id,
        }
        retries = 0
        last_logged_exception = ""
        self._assume_role()
        while True:
            if retries > self.max_kinesis_client_retries:
                raise KinesisMaxRetriesException(
                    shard_id, last_logged_exception, "Shard merging retries exceeded"
                )
            try:
                self.kinesis_client.merge_shards(**kwargs)
            except self.kinesis_client.exceptions.LimitExceededException:
                last_logged_exception = log_and_sleep(
                    f"Kinesis Limit Exceeded Exception retrying shardid: "
                    f"{shard_id} adjacent_shard_id: {adjacent_shard_id}",
                    retries,
                )
            except self.kinesis_client.exceptions.ResourceInUseException:
                last_logged_exception = log_and_sleep(
                    f"Kinesis Resource In Use Exception retrying shardid: "
                    f"{shard_id} adjacent_shard_id: {adjacent_shard_id}",
                    retries,
                )
            except self.kinesis_client.exceptions.InvalidArgumentException as e:
                if self._check_shard_already_split_or_merged(shard_id, e):
                    # disregard if the shard has already been merged or split
                    break
                last_logged_exception = log_and_sleep(
                    f"Shard Merging Error: {str(e)}", retries
                )
            except Exception as e:
                last_logged_exception = log_and_sleep(
                    f"{e} retrying shardid: {shard_id} "
                    f"adjacent_shard_id: {adjacent_shard_id}",
                    retries,
                )
            else:
                break

            retries += 1

    @time_calls
    def _wait_stream_active(self):
        kwargs = {"StreamName": self.stream_name}
        retries = 0
        self._assume_role()
        while True:
            try:
                ret = self.kinesis_client.describe_stream_summary(**kwargs)
                if ret["StreamDescriptionSummary"]["StreamStatus"] == "ACTIVE":
                    _logger.info("Stream is active")
                    return True
                else:
                    log_and_sleep(
                        f"Stream State {ret['StreamDescriptionSummary']['StreamStatus']}"
                        f" Looping waiting to become active",
                        retries,
                        max_sleep=60,
                    )
            except self.kinesis_client.exceptions.LimitExceededException:
                log_and_sleep(
                    "Kinesis Limit Exceeded Exception Describing Stream",
                    retries,
                    max_sleep=60,
                )
            except self.kinesis_client.exceptions.ResourceInUseException:
                log_and_sleep(
                    "Kinesis Resource In Use Exception Describing Stream",
                    retries,
                    max_sleep=60,
                )
            except Exception as e:
                log_and_sleep(str(e), retries, max_sleep=60)
            retries += 1

    @time_calls
    def split_pass(self):
        pass_count = 0
        shards_prev_pass = self.new_shards_map
        while True:
            start = time.time()
            pass_count += 1
            split_shards = []
            last_priority = 0
            is_low_priority_operation = False
            is_high_priority_completed = False
            initial_low_priority_operations = 0
            initial_operations = self.reshard_operations
            _logger.info(f"Starting pass {pass_count} for splits")
            self._wait_stream_active()
            #  Sleep some to give kinesis time to settle and return a good shard list.
            time.sleep(self.shard_settle_timer)
            shards = self.get_shards()
            shard_lookup = ShardLookupByPartitionKey(shards)
            # This is so on subsequent passes we can reduce the number of shards we loop through increasing performance
            shards_this_pass = shards_prev_pass
            shards_prev_pass = []
            for shard_tuple in shards_this_pass:
                if self.reshard_operations >= self.reshard_operations_limit:
                    break
                if (
                    is_low_priority_operation
                    and not is_high_priority_completed
                    and initial_operations != self.reshard_operations
                    and self.reshard_operations_low_priority
                    >= self.reshard_operations_low_priority_limit
                ):
                    shards_prev_pass = (
                        self.new_shards_map
                    )  # start from initial shard map
                    break
                (new_start_hash, new_end_hash, percent_full, priority) = shard_tuple
                (
                    shard_id,
                    start_hash,
                    end_hash,
                ) = shard_lookup.find_shard_containing_partition_key(new_end_hash)

                if (
                    self.reshard_by_priority
                    and last_priority != 0
                    and priority != last_priority
                ):
                    if not is_low_priority_operation:
                        is_low_priority_operation = True
                        if initial_operations == self.reshard_operations:
                            is_high_priority_completed = True

                    if (
                        initial_operations != self.reshard_operations
                        and self.reshard_operations_low_priority
                        >= self.reshard_operations_low_priority_limit
                    ):
                        shards_prev_pass = (
                            self.new_shards_map
                        )  # start from initial shard map
                        break

                last_priority = priority

                if new_end_hash == end_hash and (
                    new_start_hash <= start_hash or not self.split_shard_in_middle
                ):
                    _logger.debug(
                        f"Shard {shard_id} already ends on the correct boundary no split needed"
                    )
                    continue

                _logger.debug(f"We need to split shard {shard_id}")
                if shard_id in split_shards:
                    _logger.debug(
                        f"We already split shard {shard_id} once this pass, skipping until next pass"
                    )
                else:
                    if self.split_shard_in_middle:
                        _logger.debug(f"Original end hash for split is {new_end_hash}")
                        new_end_hash = self._find_new_end_hash_middle(
                            shards_this_pass, start_hash, end_hash
                        )
                        _logger.debug(
                            f"Overwritten end hash for split is {new_end_hash}"
                        )
                        _logger.debug(
                            f"Split Pass {pass_count} took {time.time() - start} seconds and found "
                            f"{len(shards_prev_pass)} shards for the next pass"
                        )
                    #  Plus 1 to pass the start hash of the second new shard
                    self._split_shard(shard_id, new_end_hash + 1)
                    split_shards.append(shard_id)
                    if initial_low_priority_operations > 0:
                        self.reshard_operations_low_priority += 1
                    self.reshard_operations += 1
                shards_prev_pass.append(shard_tuple)
            _logger.info(
                f"Split Pass {pass_count} took {time.time() - start} seconds with {len(split_shards)} splits and found "
                f"{len(shards_prev_pass)} shards for the next pass"
            )
            if len(shards_prev_pass) == 0:
                _logger.info(f"Finished splitting took {pass_count} passes")
                self.metrics_registry.gauge("split_passes").set_value(pass_count)
                break
            if pass_count >= self.reshard_passes:
                shards = self.get_shards()
                _logger.warning(
                    f"Exiting after {self.reshard_passes} split passes.  Shard State: {shards}"
                )
                self.metrics_registry.gauge("split_passes").set_value(pass_count)
                break
            if self.reshard_operations >= self.reshard_operations_limit:
                shards = self.get_shards()
                _logger.warning(
                    f"Exiting after {self.reshard_operations} reshard operations.  Shard State: {shards}"
                )
                self.metrics_registry.gauge("split_passes").set_value(pass_count)
                break
        self.split_pass_ran = True

    @time_calls
    def merge_pass(self):
        if not self.split_pass_ran:
            raise Exception("Split pass must run before Merge pass")
        pass_count = 0
        shards_prev_pass = self.new_shards_map
        while True:
            start = time.time()
            pass_count += 1
            _logger.info(f"Starting pass {pass_count} for merges")
            self._wait_stream_active()
            #  Sleep some to give kinesis time to settle and return a good shard list.
            time.sleep(self.shard_settle_timer)
            shards = self.get_shards()
            shard_lookup = ShardLookupByPartitionKey(shards)
            # This is so on subsequent passes we can reduce the number of shards we loop through increasing performance
            shards_this_pass = shards_prev_pass
            shards_prev_pass = []
            for shard_tuple in shards_this_pass:
                if self.reshard_operations >= self.reshard_operations_limit:
                    break
                new_start_hash, new_end_hash, percent_full, _ = shard_tuple
                (
                    shard_id,
                    start_hash,
                    end_hash,
                ) = shard_lookup.find_shard_containing_partition_key(new_end_hash)
                if new_start_hash == start_hash and new_end_hash == end_hash:
                    _logger.debug(f"Shard {shard_id} is good it matches desired")
                elif self.merge_shards_by_pairs:
                    shard_ids = shard_lookup.find_shards_by_range(
                        new_start_hash, new_end_hash
                    )
                    _logger.debug(
                        f"Shards between {new_start_hash} and {new_end_hash} are: {shard_ids}"
                    )
                    for (shardLeft, shardRight) in [
                        shard_ids[i : i + 2] for i in range(0, len(shard_ids) - 1, 2)
                    ]:
                        _logger.debug(f"Merge shards {shardLeft}:{shardRight}")
                        self._merge_shards(shardLeft, shardRight)
                        self.reshard_operations += 1
                    if len(shard_ids) >= 2:
                        shards_prev_pass.append(shard_tuple)
                else:
                    # Find Adjacent shard
                    (
                        next_shard_id,
                        next_start_hash,
                        next_end_hash,
                    ) = shard_lookup.find_shard_containing_partition_key(end_hash + 1)
                    (
                        prev_shard_id,
                        prev_start_hash,
                        prev_end_hash,
                    ) = shard_lookup.find_shard_containing_partition_key(start_hash - 1)
                    if next_shard_id is not None and next_end_hash <= new_end_hash:
                        _logger.debug(
                            f"Merge with next shard {shard_id}:{next_shard_id}"
                        )
                        self._merge_shards(shard_id, next_shard_id)
                        self.reshard_operations += 1
                        shards_prev_pass.append(shard_tuple)
                    elif (
                        prev_shard_id is not None and prev_start_hash >= new_start_hash
                    ):
                        _logger.debug(
                            f"Merge with prev shard {shard_id}:{prev_shard_id}"
                        )
                        self._merge_shards(prev_shard_id, shard_id)
                        shards_prev_pass.append(shard_tuple)
                        self.reshard_operations += 1
                    else:
                        _logger.warning(
                            "Attempting to merge where start and end hashes are out of range, "
                            "Likely we have only one shard"
                        )
            _logger.info(
                f"Merge Pass {pass_count} took {time.time() - start} seconds and found "
                f"{len(shards_prev_pass)} shards for the next pass"
            )
            if self.reshard_operations >= self.reshard_operations_limit:
                shards = self.get_shards()
                _logger.warning(
                    f"Exiting after {self.reshard_operations} reshard operations.  Shard State: {shards}"
                )
                self.metrics_registry.gauge("merge_passes").set_value(pass_count)
                break
            if len(shards_prev_pass) == 0:
                _logger.info(f"Finished merging took {pass_count} passes")
                self.metrics_registry.gauge("merge_passes").set_value(pass_count)
                break
            if pass_count >= self.reshard_passes:
                shards = self.get_shards()
                _logger.warning(
                    f"Exiting after {self.reshard_passes} merge passes.  Shard State: {shards}"
                )
                self.metrics_registry.gauge("merge_passes").set_value(pass_count)
                break

    @time_calls
    def reshard(self, disable_shard_merge=False, disable_shard_split=False):
        self.reshard_operations = 0
        self.reshard_operations_low_priority = 0
        self.build_new_stream_shards()
        if not disable_shard_split:
            self.split_pass()
        if not disable_shard_merge:
            self.merge_pass()

    def shard_map_output(self, admin_bucket, shard_map_key_prefix):
        start = time.time()
        shards = self.get_active_shards()
        self.metrics_registry.gauge("active_shards").set_value(len(shards))
        shard_map = ""
        partition_key_sorted_list = sorted(list(self.partition_keys.keys()))
        for shard in sorted(
            shards, key=lambda i: int(i["HashKeyRange"]["StartingHashKey"])
        ):
            shard_map += (
                f"{shard['ShardId']}\t{shard['HashKeyRange']['StartingHashKey']}\n"
            )
            first_key_location = bisect_left(
                partition_key_sorted_list, int(shard["HashKeyRange"]["StartingHashKey"])
            )
            last_key_location = bisect_right(
                partition_key_sorted_list, int(shard["HashKeyRange"]["EndingHashKey"])
            )
            keys = partition_key_sorted_list[first_key_location:last_key_location]
            total = 0
            for par_key in keys:
                total += self.partition_metadata[par_key]["percent"]
                shard_map += (
                    f"\t{par_key}\t{self.partition_metadata[par_key]['name']}\t"
                    f"{self.partition_metadata[par_key]['percent']}\t{total}\n"
                )

        _logger.info(
            f"Saving Shard Map to S3 s3://{admin_bucket}/{shard_map_key_prefix}latest.txt"
        )

        timestamp = datetime.now().isoformat()
        self.s3_client.put_object(
            Body=shard_map, Bucket=admin_bucket, Key=f"{shard_map_key_prefix}latest.txt"
        )
        self.s3_client.put_object(
            Body=shard_map,
            Bucket=admin_bucket,
            Key=f"{shard_map_key_prefix}{timestamp}.txt",
        )
        _logger.info(
            f"Took {time.time() - start} seconds to create and write shard map."
        )

    def _get_max_partition_priority(self, partition_keys):
        return max(
            [
                self.partition_metadata[k]["priority"]
                for k in partition_keys
                if k in self.partition_metadata
            ],
            default=0,
        )

    def _sort_stream_shards_by_priority(self, stream_shards):
        new_stream_shards = []
        # sort by priority first and start hash range second
        for shard in sorted(stream_shards, key=lambda i: (i[3], i[0]), reverse=True):
            new_stream_shards.append(shard)
        return new_stream_shards


class KinesisMockShardManager(ShardManager):
    def __init__(self, partition_keys, partition_metadata, config, shard_data=None):
        self.shard_data = shard_data
        self.test_shard_id = self._get_next_shard_id()
        super().__init__(partition_keys, partition_metadata, config)

    @time_calls
    def get_shards(self):
        return self.shard_data

    def _wait_stream_active(self):
        return True

    @time_calls
    def _split_shard(self, shard_id, hash_key):
        self.shard_splits.append((shard_id, hash_key))
        added_shards = []
        for index, shard in enumerate(self.shard_data, 0):
            if shard["ShardId"] == shard_id:
                self.test_shard_id += 1
                self.shard_data[index]["SequenceNumberRange"][
                    "EndingSequenceNumber"
                ] = "901283792038740928309873"
                added_shards.append(
                    {
                        "ShardId": f"ShardId-{self.test_shard_id:012}",
                        "ParentShardId": shard_id,
                        "HashKeyRange": {
                            "StartingHashKey": shard["HashKeyRange"]["StartingHashKey"],
                            "EndingHashKey": str(hash_key - 1),
                        },
                        "SequenceNumberRange": {
                            "StartingSequenceNumber": "49596888501586614134748550254232887619159170724774543394"
                        },
                    }
                )
                self.test_shard_id += 1
                added_shards.append(
                    {
                        "ShardId": f"ShardId-{self.test_shard_id:012}",
                        "ParentShardId": shard_id,
                        "HashKeyRange": {
                            "StartingHashKey": str(hash_key),
                            "EndingHashKey": shard["HashKeyRange"]["EndingHashKey"],
                        },
                        "SequenceNumberRange": {
                            "StartingSequenceNumber": "49596888501586614134748550254232887619159170724774543394"
                        },
                    }
                )
        self.shard_data.extend(added_shards)

    @time_calls
    def _merge_shards(self, shard_id, adjacent_shard_id):
        self.shard_merges.append((shard_id, adjacent_shard_id))
        added_shards = []
        start_hashes = []
        end_hashes = []
        for index, shard in enumerate(self.shard_data, 0):
            if shard["ShardId"] == shard_id or shard["ShardId"] == adjacent_shard_id:
                self.shard_data[index]["SequenceNumberRange"][
                    "EndingSequenceNumber"
                ] = "901283792038740928309873"
                start_hashes.append(int(shard["HashKeyRange"]["StartingHashKey"]))
                end_hashes.append(int(shard["HashKeyRange"]["EndingHashKey"]))
        start_hash = min(start_hashes)
        end_hash = max(end_hashes)
        self.test_shard_id += 1
        added_shards.append(
            {
                "ShardId": f"ShardId-{self.test_shard_id:012}",
                "ParentShardId": shard_id,
                "HashKeyRange": {
                    "StartingHashKey": str(start_hash),
                    "EndingHashKey": str(end_hash),
                },
                "SequenceNumberRange": {
                    "StartingSequenceNumber": "49596888501586614134748550254232887619159170724774543394"
                },
            }
        )
        self.shard_data.extend(added_shards)

    def _get_next_shard_id(self):
        prev_shard_num = 0
        for shard in self.shard_data:
            shard_num = int(shard["ShardId"].split("-")[1])
            if shard_num > prev_shard_num:
                prev_shard_num = shard_num
        return prev_shard_num


class ShardLookupByPartitionKey:
    MAX_KEY = int("0xffffffffffffffffffffffffffffffff", 0)

    @time_calls
    def __init__(self, shards):
        self.ordered_lookup = []
        self.shard_lookup_by_end_hash = {}
        for shard in shards:
            # only lookup active shards so eliminate any with an EndingSequenceNumber
            if "EndingSequenceNumber" not in shard["SequenceNumberRange"]:
                end_hash = int(shard["HashKeyRange"]["EndingHashKey"])
                self.ordered_lookup.append(end_hash)
                self.shard_lookup_by_end_hash[end_hash] = shard
        self.ordered_lookup.sort()

    @time_calls
    def find_shard_containing_partition_key(self, key):
        if not (0 <= key <= self.MAX_KEY):
            return None, None, None
        end_key_index = bisect.bisect_left(self.ordered_lookup, key)
        end_hash = self.ordered_lookup[end_key_index]
        shard_id = self.shard_lookup_by_end_hash[end_hash]["ShardId"]
        start_hash = int(
            self.shard_lookup_by_end_hash[end_hash]["HashKeyRange"]["StartingHashKey"]
        )
        return shard_id, start_hash, end_hash

    @time_calls
    def find_shards_by_range(self, start, end):
        if not (0 <= start < end <= self.MAX_KEY):
            return []
        shard_ids = []
        start_key_index = bisect.bisect_right(self.ordered_lookup, start)
        end_key_index = bisect.bisect_right(self.ordered_lookup, end)
        for end_hash in self.ordered_lookup[start_key_index:end_key_index]:
            start_hash = int(
                self.shard_lookup_by_end_hash[end_hash]["HashKeyRange"][
                    "StartingHashKey"
                ]
            )
            if start_hash >= start and end_hash <= end:
                shard_ids.append(self.shard_lookup_by_end_hash[end_hash]["ShardId"])
        return shard_ids
